/* Copyright 2013-2017 Sathya Laufer
 *
 * Homegear is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * Homegear is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Homegear.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU Lesser General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
*/

#include "Auth.h"
#include "../GD/GD.h"

namespace Rpc
{
Auth::Auth()
{
	_socket =  std::shared_ptr<BaseLib::TcpSocket>(new BaseLib::TcpSocket(GD::bl.get()));
}

Auth::Auth(std::shared_ptr<BaseLib::TcpSocket>& socket, std::vector<std::string>& validUsers)
{
	_socket = socket;
	_validUsers = validUsers;
	_initialized = true;
	_rpcEncoder = std::shared_ptr<BaseLib::Rpc::RpcEncoder>(new BaseLib::Rpc::RpcEncoder(GD::bl.get()));
	_jsonDecoder = std::shared_ptr<BaseLib::Rpc::JsonDecoder>(new BaseLib::Rpc::JsonDecoder(GD::bl.get()));
}

Auth::Auth(std::shared_ptr<BaseLib::TcpSocket>& socket, std::string userName, std::string password)
{
	_socket = socket;
	_password = password;
	if(!_userName.empty() && !_password.empty()) _initialized = true;
}

std::pair<std::string, std::string> Auth::basicClient()
{
	if(!_initialized) throw AuthException("Not initialized.");
	if(_basicAuthString.second.empty())
	{
		_basicAuthString.first = "Authorization";
		std::string credentials = _userName + ":" + _password;
		std::string encodedData;
		BaseLib::Base64::encode(credentials, encodedData);
		_basicAuthString.second = "Basic " + encodedData;
	}
	return _basicAuthString;
}

void Auth::sendBasicUnauthorized(bool binary)
{
	if(binary)
	{
		if(_basicUnauthBinaryHeader.empty())
		{
			BaseLib::PVariable error = BaseLib::Variable::createError(-32603, "Unauthorized");
			_rpcEncoder->encodeResponse(error, _basicUnauthBinaryHeader);
		}
		try
		{
			_socket->proofwrite(_basicUnauthBinaryHeader);
		}
		catch(const BaseLib::SocketOperationException& ex)
		{
			throw AuthException("Authorization failed because of socket exception: " + ex.what());
		}
	}
	else
	{
		if(_basicUnauthHTTPHeader.empty())
		{
			std::string header;
			header.append("HTTP/1.1 401 Unauthorized\r\n");
			header.append("Connection: Close\r\n");
			std::string content("<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01//EN\" \"http://www.w3.org/TR/html4/strict.dtd\"><html><head><meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\"><title>Unauthorized</title></head><body>Unauthorized</body></html>");
			header.append("Content-Length: " + std::to_string(content.size()) + "\r\n\r\n");
			header.append(content);
			_basicUnauthHTTPHeader.insert(_basicUnauthHTTPHeader.begin(), header.begin(), header.end());
		}
		try
		{
			_socket->proofwrite(_basicUnauthHTTPHeader);
		}
		catch(const BaseLib::SocketOperationException& ex)
		{
			throw AuthException("Authorization failed because of socket exception: " + ex.what());
		}
	}
}

bool Auth::basicServer(std::shared_ptr<BaseLib::Rpc::RpcHeader>& binaryHeader, std::string& userName, BaseLib::Security::PAcls& acls)
{
	userName = "";
	if(!acls) acls = std::make_shared<BaseLib::Security::Acls>(GD::bl.get(), -1);
	acls->clear();
	if(!_initialized) throw AuthException("Not initialized.");
	if(!binaryHeader) throw AuthException("Header is nullptr.");
	if(binaryHeader->authorization.empty())
	{
		sendBasicUnauthorized(true);
		throw AuthException("No header field \"Authorization\"");
	}
	std::pair<std::string, std::string> authData = BaseLib::HelperFunctions::splitLast(binaryHeader->authorization, ' ');
	BaseLib::HelperFunctions::toLower(authData.first);
	if(authData.first != "basic")
	{
		sendBasicUnauthorized(true);
		throw AuthException("Authorization type is not basic but: " + authData.first);
	}
	std::string decodedData;
	BaseLib::Base64::decode(authData.second, decodedData);
	std::pair<std::string, std::string> credentials = BaseLib::HelperFunctions::splitLast(decodedData, ':');
	BaseLib::HelperFunctions::toLower(credentials.first);
	if(credentials.first.empty()) throw AuthException("User name is empty.");
	if(std::find(_validUsers.begin(), _validUsers.end(), credentials.first) == _validUsers.end())
	{
		sendBasicUnauthorized(true);
		throw AuthException("User name " + credentials.first + " is not in the list of valid users in /etc/homegear/rpcservers.conf.");
	}
	if(User::verify(credentials.first, credentials.second))
	{
		userName = credentials.first;
		if(!acls->fromUser(userName))
		{
			_userName = "";
			throw AuthException("Could not set ACLs.");
		}
		return true;
	}
	sendBasicUnauthorized(true);
	return false;
}

bool Auth::basicServer(BaseLib::Http& httpPacket, std::string& userName, BaseLib::Security::PAcls& acls)
{
	userName = "";
	if(!acls) acls = std::make_shared<BaseLib::Security::Acls>(GD::bl.get(), -1);
	acls->clear();
	if(!_initialized) throw AuthException("Not initialized.");
	_http.reset();
	uint32_t bufferLength = 1024;
	char buffer[bufferLength + 1];
	if(httpPacket.getHeader().authorization.empty())
	{
		if(_basicAuthHTTPHeader.empty())
		{
			_basicAuthHTTPHeader.append("HTTP/1.1 401 Authorization Required\r\n");
			_basicAuthHTTPHeader.append("WWW-Authenticate: Basic realm=\"Authentication Required\"\r\n");
			_basicAuthHTTPHeader.append("Connection: Keep-Alive\r\n");
			std::string content("<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01//EN\" \"http://www.w3.org/TR/html4/strict.dtd\"><html><head><meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\"><title>Authorization Required</title></head><body>Authorization Required</body></html>");
			_basicAuthHTTPHeader.append("Content-Length: " + std::to_string(content.size()) + "\r\n\r\n");
			_basicAuthHTTPHeader.append(content);
		}
		std::shared_ptr<std::vector<char>> data(new std::vector<char>());
		data->insert(data->begin(), _basicAuthHTTPHeader.begin(), _basicAuthHTTPHeader.end());
		try
		{
			_socket->proofwrite(data);
			int32_t bytesRead = _socket->proofread(buffer, bufferLength);
			//Some clients send only one byte in the first packet
			if(bytesRead == 1) bytesRead += _socket->proofread(&buffer[1], bufferLength - 1);
			buffer[bytesRead] = '\0';
			try
			{
				_http.process(buffer, bufferLength);
			}
			catch(BaseLib::HttpException& ex)
			{
				throw AuthException("Authorization failed because of HTTP exception: " + ex.what());
			}
		}
		catch(const BaseLib::SocketOperationException& ex)
		{
			throw AuthException("Authorization failed because of socket exception: " + ex.what());
		}
	}
	else _http = httpPacket;
	if(_http.getHeader().authorization.empty())
	{
		sendBasicUnauthorized(false);
		throw AuthException("No header field \"Authorization\"");
	}
	std::pair<std::string, std::string> authData = BaseLib::HelperFunctions::splitLast(_http.getHeader().authorization, ' ');
	BaseLib::HelperFunctions::toLower(authData.first);
	if(authData.first != "basic")
	{
		sendBasicUnauthorized(false);
		throw AuthException("Authorization type is not basic but: " + authData.first);
	}
	std::string decodedData;
	BaseLib::Base64::decode(authData.second, decodedData);
	std::pair<std::string, std::string> credentials = BaseLib::HelperFunctions::splitLast(decodedData, ':');
	BaseLib::HelperFunctions::toLower(credentials.first);
	if(credentials.first.empty()) throw AuthException("User name is empty.");
	if(std::find(_validUsers.begin(), _validUsers.end(), credentials.first) == _validUsers.end())
	{
		sendBasicUnauthorized(false);
		throw AuthException("User name " + credentials.first + " is not in the list of valid users in /etc/homegear/rpcservers.conf.");
	}
	if(User::verify(credentials.first, credentials.second))
	{
		userName = credentials.first;
		if(!acls->fromUser(userName))
		{
			_userName = "";
			throw AuthException("Could not set ACLs.");
		}
		return true;
	}
	sendBasicUnauthorized(false);
	return false;
}

bool Auth::basicServer(BaseLib::WebSocket& webSocket, std::string& userName, BaseLib::Security::PAcls& acls)
{
	userName = "";
	if(!acls) acls = std::make_shared<BaseLib::Security::Acls>(GD::bl.get(), -1);
	acls->clear();
	if(!_initialized) throw AuthException("Not initialized.");
	if(webSocket.getContent().empty())
	{
		sendWebSocketUnauthorized("No data received.");
		return false;
	}
	BaseLib::PVariable variable;
	try
	{
		variable = _jsonDecoder->decode(webSocket.getContent());
	}
	catch(BaseLib::Rpc::JsonDecoderException& ex)
	{
		sendWebSocketUnauthorized("Error decoding json (is packet in json format?).");
		return false;
	}
	if(variable->type != BaseLib::VariableType::tStruct)
	{
		sendWebSocketUnauthorized("Received data is no json object.");
		return false;
	}
	if(variable->structValue->find("user") == variable->structValue->end() || variable->structValue->find("password") == variable->structValue->end())
	{
		sendWebSocketUnauthorized("Either 'user' or 'password' is not specified.");
		return false;
	}
	BaseLib::HelperFunctions::toLower(variable->structValue->at("user")->stringValue);
	if(variable->structValue->at("user")->stringValue.empty()) throw AuthException("User name is empty.");
	if(User::verify(variable->structValue->at("user")->stringValue, variable->structValue->at("password")->stringValue))
	{
		if(!acls->fromUser(variable->structValue->at("user")->stringValue))
		{
			_userName = "";
			throw AuthException("Could not set ACLs.");
		}
		sendWebSocketAuthorized();
		return true;
	}
	sendWebSocketUnauthorized("Wrong user name or password.");
	return false;
}

bool Auth::sessionServer(BaseLib::WebSocket& webSocket, std::string& userName, BaseLib::Security::PAcls& acls)
{
	userName = "";
	if(!acls) acls = std::make_shared<BaseLib::Security::Acls>(GD::bl.get(), -1);
	acls->clear();
	if(!_initialized) throw AuthException("Not initialized.");
	if(webSocket.getContent().empty())
	{
		sendWebSocketUnauthorized("No data received.");
		return false;
	}
#ifndef NO_SCRIPTENGINE
	BaseLib::PVariable variable;
	try
	{
		variable = _jsonDecoder->decode(webSocket.getContent());
	}
	catch(BaseLib::Rpc::JsonDecoderException& ex)
	{
		sendWebSocketUnauthorized("Error decoding json (is packet in json format?).");
		return false;
	}
	if(variable->type != BaseLib::VariableType::tStruct)
	{
		sendWebSocketUnauthorized("Received data is no json object.");
		return false;
	}
	if(variable->structValue->find("user") != variable->structValue->end())
	{
		std::string userName = GD::scriptEngineServer->checkSessionId(variable->structValue->at("user")->stringValue);
		BaseLib::HelperFunctions::toLower(userName);
		if(!userName.empty())
		{
			if(!acls->fromUser(userName))
			{
				throw AuthException("Could not set ACLs (user: " + userName + ")");
			}
			sendWebSocketAuthorized();
			return true;
		}
		else
		{
			sendWebSocketUnauthorized("Session id is invalid.");
			return false;
		}
	}
	else
	{
		sendWebSocketUnauthorized("No session id specified.");
		return false;
	}
#else
	sendWebSocketUnauthorized("Homegear is compiled without script engine. Session authentication is not possible.");
	throw AuthException("Homegear is compiled without script engine. Session authentication is not possible.");
	return false;
#endif
}

void Auth::sendWebSocketAuthorized()
{
	std::vector <char> output;
	std::string json("{\"auth\":\"success\"}");
	std::vector<char> data(&json[0], &json[0] + json.size());
	BaseLib::WebSocket::encode(data, BaseLib::WebSocket::Header::Opcode::text, output);
	_socket->proofwrite(output);
}

void Auth::sendWebSocketUnauthorized(std::string reason)
{
	std::vector <char> output;
	std::string json("{\"auth\":\"failure\",\"reason\":\"" + reason + "\"}");
	std::vector<char> data(json.data(), json.data() + json.size());
	BaseLib::WebSocket::encode(data, BaseLib::WebSocket::Header::Opcode::text, output);
	_socket->proofwrite(output);
}

} /* namespace Rpc */
